﻿using System;
using System.Collections.Generic;
using System.Xml;
using System.IO;

namespace STFXmlParser
{
    class XmlParser
    {
        public XmlParser(ref PageArray pa, byte[] data)
        {
            buffer = new MemoryStream(data);
            ElementName = "";
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreWhitespace = true;
            settings.IgnoreComments = true;
            settings.DtdProcessing = DtdProcessing.Ignore;

            reader = XmlReader.Create(buffer, settings);

            pageArray = pa;
        }

        public bool Parse()
        {
            Dictionary<string,string> Attribs = new Dictionary<string, string>();

            while (reader.Read())
            {
                if (reader.IsStartElement())
                {
                    if (reader.HasAttributes)
                    {
                        while (reader.MoveToNextAttribute())
                        {
                            Attribs.Add(reader.Name, reader.Value);
                        }

                        reader.MoveToElement();
                    }

                    ProcessStartElement(reader.Name, Attribs);
                    Attribs.Clear();
                    
                    if (reader.IsEmptyElement )
                        ProcessEndElement(reader.Name);
                }
                else if (reader.NodeType == XmlNodeType.EndElement)
                {
                    ProcessEndElement(reader.Name);
                }
                else if (reader.NodeType == XmlNodeType.CDATA || reader.NodeType == XmlNodeType.Text)
                {
                    ProcessCharacters(reader.ReadContentAsString());

                    if ( reader.NodeType == XmlNodeType.EndElement )
                        ProcessEndElement(reader.Name);

                }
                
            }

            // Set the list that contains all page numbers that have boxes on them.
            pageArray.SetBoxedPageList();

            return true;
        }

        private void ProcessStartElement(String name, Dictionary<string, string> Attributes)
        {
            //Console.WriteLine(name);
            String Value;

            if (name.ToLower() == "form")
            {
                foreach (KeyValuePair<string, string> kvp in Attributes)
                {
                    switch (kvp.Key)
                    {
                        case "ref":
                            pageArray.STFName = kvp.Value;
                            break;
                    }
                }
            }
            else if (name.ToLower() == "page")
            {
                if (Attributes.Count > 0)
                {
                    if (Attributes.TryGetValue("ref", out Value))
                    {
                        Value = Value.Substring(2);
                        CurrentPage = Convert.ToInt32(Value);
                    }
                }
            }
            else if (name.ToLower() == "pdf417" || name.ToLower() == "code39" || name.ToLower() == "int2of5" || name.ToLower() == "qrcode")
            {
                BarcodeField ptr = new BarcodeField();

                foreach (KeyValuePair<string, string> kvp in Attributes)
                {
                    switch (kvp.Key)
                    {
                        case "ref":
                            ptr.ref_name = kvp.Value;
                            break;
                    }
                }

                if (ptr.ref_name.Length > 0)
                {
                    pageArray.AddPageField(CurrentPage - 1, ptr);
                    pageArray.AddBoxPage(name.ToLower(), (uint)CurrentPage);
                }
            }
            else if (name.ToLower() == "text")
            {
                TextField ptr = new TextField();

                foreach (KeyValuePair<string, string> kvp in Attributes)
                {
                    switch (kvp.Key)
                    {
                        case "ref":
                            ptr.ref_name = kvp.Value;
                            break;

                        case "justification":
                            {
                                switch (kvp.Value)
                                {
                                    case "left":
                                        ptr.just = TextField.justification.left;
                                        break;
                                    case "right":
                                        ptr.just = TextField.justification.right;
                                        break;
                                    case "center":
                                        ptr.just = TextField.justification.center;
                                        break;
                                    case "decimal":
                                        ptr.just = TextField.justification.dec;
                                        break;
                                    case "decimal_no_point":
                                        ptr.just = TextField.justification.dec_np;
                                        break;
                                }
                            }
                            break;

                        case "multiline":
                            if (kvp.Value == "true")
                                ptr.multiline = true;

                            break;

                        case "visable":
                            if (kvp.Value == "false")
                                ptr.visible = false;
                            break;

                        case "editable":
                            if (kvp.Value == "false")
                                ptr.editable = false;
                            break;
                        
                        case "checkbox":
                            if (kvp.Value == "true")
                                ptr.checkbox = true;
                            break;

                        case "case":
                            if (kvp.Value == "upper")
                                ptr.tc = TextField.text_case.upper;
                            else if (kvp.Value == "lower")
                                ptr.tc = TextField.text_case.lower;
                            else
                                ptr.tc = TextField.text_case.mixed;

                            break;

                        case "vertical_alignment":
                            if (kvp.Value == "bottom")
                                ptr.va = TextField.vert_align.bottom;
                            else if (kvp.Value == "baseline")
                                ptr.va = TextField.vert_align.baseline;
                            else
                                ptr.va = TextField.vert_align.bottom;

                            break;

                        case "link":
                            ptr.link_page = Convert.ToUInt32(kvp.Value);
                            break;

                        case "insert_commas":
                            {
                                if (kvp.Value == "true")
                                    ptr.insert_commas = true;
                                else if (kvp.Value == "false")
                                    ptr.insert_commas = false;
                                else if (kvp.Value == "true_locked")
                                {
                                    ptr.insert_commas = true;
                                    ptr.commas_locked = true;
                                }
                                else if (kvp.Value == "false_locked")
                                {
                                    ptr.insert_commas = false;
                                    ptr.commas_locked = true;
                                }
                            }
                            break;

                        case "negatives_in_parenthesis":
                            {
                                if (kvp.Value == "true")
                                    ptr.negs_in_parens = true;
                                else if (kvp.Value == "false")
                                    ptr.negs_in_parens = false;
                                else if (kvp.Value == "true_locked")
                                {
                                    ptr.negs_in_parens = true;
                                    ptr.negs_locked = true;
                                }
                                else if (kvp.Value == "false_locked")
                                {
                                    ptr.negs_in_parens = false;
                                    ptr.negs_locked = true;
                                }
                            }
                            break;
                    }

                } // end foreach

                if (ptr.ref_name.Length > 0)
                {
                    pageArray.AddPageField(CurrentPage - 1, ptr);

                    if (ptr.checkbox == true)
                    {
                        pageArray.AddBoxPage("checkbox", (uint)CurrentPage);
                    }
                }
            }
            else if (name.ToLower() == "cid")
            {
                CidField ptr = new CidField();

                foreach (KeyValuePair<string, string> kvp in Attributes)
                {
                    switch (kvp.Key)
                    {
                        case "ref":
                            ptr.ref_name = kvp.Value;
                            break;

                        // Just keeping the "ref" name above as we are keeping
                        // track of cid boxes in the documents code file where
                        // we use cid1_1 to refer to a cid box.
                        //case "cid_ref":
                        //    ptr.ref_name += (':' + kvp.Value);
                        //    break;

                        case "justification":
                            {
                                switch (kvp.Value)
                                {
                                    case "left":
                                        ptr.just = CidField.justification.left;
                                        break;
                                    case "right":
                                        ptr.just = CidField.justification.right;
                                        break;
                                    case "center":
                                        ptr.just = CidField.justification.center;
                                        break;
                                }
                            }
                            break;

                        case "visable":
                            if (kvp.Value == "false")
                                ptr.visible = false;
                            break;

                        case "rotate":
                            if (kvp.Value != "0")
                                ptr.rotation = kvp.Value;
                            break;
                    }

                } // end foreach

                if (ptr.ref_name.Length > 0)
                    pageArray.AddPageField(CurrentPage - 1, ptr);
            }
            else if ((name.ToLower() == "data") || (name.ToLower() == "cid_value"))
            {

                ElementName = name.ToLower();

            }
            else if ((name.ToLower() == "bbox") || (name.ToLower() == "cid_bbox"))
            {
                Field ptr = pageArray.GetBackPageField(CurrentPage - 1);
                float left = 0;
                float right = 0;
                float bottom = 0;
                float top = 0;

                foreach (KeyValuePair<string, string> kvp in Attributes)
                {

                    switch (kvp.Key)
                    {
                        case "left":
                            left = (float)Convert.ToDouble(kvp.Value);
                            break;

                        case "right":
                            right = (float)Convert.ToDouble(kvp.Value);
                            break;

                        case "top":
                            top = (float)Convert.ToDouble(kvp.Value);
                            break;

                        case "bottom":
                            bottom = (float)Convert.ToDouble(kvp.Value);
                            break;
                    }

                }

                ptr.bbox = new FloatRect(left, top, right, bottom);
            }
            else if (name.ToLower() == "font")
            {
                TextField ptr = (TextField)pageArray.GetBackPageField(CurrentPage - 1);

                if (ptr != null)
                {
                    ElementName = name.ToLower();

                    foreach (KeyValuePair<string, string> kvp in Attributes)
                    {

                        if (kvp.Key == "size")
                            ptr.font_size = Convert.ToUInt32(kvp.Value);
                        else if (kvp.Key == "style")
                        {
                            if (kvp.Value == "bold")
                                ptr.font_bold = true;
                            else if (kvp.Value == "italic")
                                ptr.font_italic = true;
                            else if (kvp.Value == "bolditalic")
                            {
                                ptr.font_bold = true;
                                ptr.font_italic = true;
                            }
                        }
                        else if (kvp.Key == "locked")
                        {
                            if (kvp.Value == "true")
                                ptr.font_locked = true;
                        }
                    }
                }
            }
            else if (name.ToLower() == "cid_font")
            {
                CidField ptr = (CidField)pageArray.GetBackPageField(CurrentPage - 1);

                if (ptr != null)
                {
                    ElementName = name.ToLower();

                    foreach (KeyValuePair<string, string> kvp in Attributes)
                    {

                        if (kvp.Key == "size")
                            ptr.font_size = Convert.ToUInt32(kvp.Value);
                        else if (kvp.Key == "style")
                        {
                            if (kvp.Value == "bold")
                                ptr.font_bold = true;
                            else if (kvp.Value == "italic")
                                ptr.font_italic = true;
                            else if (kvp.Value == "bolditalic")
                            {
                                ptr.font_bold = true;
                                ptr.font_italic = true;
                            }
                        }
                    }
                }
            }
            else if (name.ToLower() == "description")
                ElementName = name.ToLower();
            else if (name.ToLower() == "regex")
                ElementName = name.ToLower();
        }

        private void ProcessEndElement( String name )
        {
            //Console.WriteLine("#" + name);
        }

        private void ProcessCharacters( String characters )
        {
            //Console.WriteLine(characters);

            if (CurrentPage > 0 && ElementName.Length > 0)
	        {
		        Field pField = pageArray.GetBackPageField(CurrentPage - 1);

		        if ((ElementName == "data") || (ElementName == "cid_value"))
		        {
			
                    pField.data = characters;
                }
                else if (ElementName == "cid_font")
                {
                    CidField ptr = (CidField)pField;

                    if (ptr != null)
                    {
                        if (ElementName == "cid_font")
                        {
                            ptr.font_name = characters;
                        }
                    }
                }
	            else
                {
                    TextField ptr = (TextField)pField;
		
		            if (ptr != null)
                    {
			            if (ElementName == "font")
                        {
                            ptr.font_name = characters;
                        }
                        else if (ElementName == "description")
                        {
                            ptr.validation_description = characters;
                        }
                        else if (ElementName == "regex")
                        {
                            ptr.validation = characters;
                        }
                    }
			    }
		    }
        }
   
        private XmlReader reader;
        private MemoryStream buffer;
        private PageArray pageArray;
        private int CurrentPage;
        private String ElementName;
    }
}
